(*  Title:      HOL/Tools/BNF/bnf_nu_fp_sugar_tactics.ML
    Author:     Jasmin Blanchette, Inria, LORIA, MPII
    Copyright   2016

Tactics for sugared nonuniform (co)datatype constructions.
*)

signature BNF_NU_FP_SUGAR_TACTICS =
sig
  val mk_combined_type_definition_tac: Proof.context -> thm list -> thm list -> thm -> thm -> thm ->
    thm -> thm list -> thm list -> tactic
  val mk_nonuniform_induct_P_ctor0_tac: Proof.context -> int -> int -> int list -> int list list ->
    thm list -> thm -> thm list -> thm list -> thm list -> thm list -> tactic
  val mk_nonuniform_induct_P_ctor_tac: Proof.context -> thm list -> thm list -> thm list ->
    thm list -> thm list list -> thm list list -> thm list list -> thm -> tactic
  val mk_nonuniform_ind0_tac: Proof.context -> thm -> thm -> tactic
  val mk_nonuniform_induct_Q_dtor0_tac: Proof.context -> thm list -> thm list -> thm list ->
    thm list -> thm list -> thm list -> thm list -> int -> int -> int -> thm -> thm -> thm -> thm ->
    thm list -> thm list list -> thm list list -> tactic
  val mk_nonuniform_induct_Q_dtor_tac: Proof.context -> int -> int list -> thm list -> thm list ->
    thm list -> thm list -> thm list -> thm list -> thm list -> thm -> thm list -> tactic
end;

structure BNF_NU_FP_Sugar_Tactics : BNF_NU_FP_SUGAR_TACTICS =
struct

open BNF_Util
open BNF_Tactics
open BNF_FP_Util
open BNF_FP_Def_Sugar_Tactics

fun mk_combined_type_definition_tac ctxt live_arg_abs_inverses live_arg_rep_inverses pre_abs_inverse
    pre_rep_inverse pre_map_ident0 pre_map_comp fp_map_ident0s fp_map_comps =
  unfold_thms_tac ctxt (pre_abs_inverse :: pre_rep_inverse :: pre_map_ident0 :: pre_map_comp ::
    live_arg_abs_inverses @ live_arg_rep_inverses @ fp_map_ident0s @ fp_map_comps @
    @{thms type_definition_def o_def[abs_def] id_def}) THEN
  REPEAT_DETERM (SOMEGOAL (resolve_tac ctxt @{thms allI conjI impI UNIV_I refl}));

fun mk_nonuniform_induct_P_ctor0_tac ctxt nn' n ms kkss ctr_defs pre_type_definition P_ctrs
    abs_inverses set_maps pre_set_defs =
  HEADGOAL (Method.insert_tac ctxt P_ctrs THEN' rotate_tac (~ n)) THEN
  unfold_thms_tac ctxt (o_apply :: ctr_defs) THEN
  HEADGOAL (rtac ctxt (mk_absumprodE pre_type_definition ms)) THEN
  co_induct_inst_as_projs_tac ctxt 0 THEN
  EVERY (@{map 3} (mk_induct_discharge_prem_tac ctxt nn' n abs_inverses abs_inverses set_maps
      pre_set_defs)
    ms (1 upto n) kkss);

fun mk_nonuniform_induct_P_ctor_prem_tac ctxt fp_map_comps fp_map_idents fp_inj_maps P_mono
    live_arg_abs_injects live_arg_rep_injects live_arg_abs_inverses =
  HEADGOAL (rtac ctxt (unfold_thms ctxt @{thms inj_on_def} P_mono RS mp) THEN'
    EVERY' (map (fn live_arg_abs_inject =>
        rtac ctxt ballI THEN' rtac ctxt ballI THEN' rtac ctxt impI THEN'
        etac ctxt (live_arg_abs_inject RS iffD1))
      live_arg_abs_injects) THEN'
    dtac ctxt meta_spec THEN' etac ctxt meta_mp THEN'
    rtac ctxt @{thm inj_image_mem_iff[THEN iffD1]} THEN'
    resolve_tac ctxt fp_inj_maps THEN'
    EVERY' (map (fn live_arg_rep_inject =>
        SELECT_GOAL (unfold_thms_tac ctxt @{thms inj_on_def}) THEN'
        rtac ctxt ballI THEN' rtac ctxt ballI THEN' rtac ctxt impI THEN'
        etac ctxt (live_arg_rep_inject RS iffD1))
      live_arg_rep_injects) THEN'
    SELECT_GOAL (unfold_thms_tac ctxt (@{thm o_def} :: live_arg_abs_inverses @ fp_map_comps @
      fp_map_idents)) THEN'
    assume_tac ctxt);

fun mk_nonuniform_induct_P_ctor_tac ctxt fp_map_comps fp_map_idents fp_inj_maps P_monos
    live_arg_abs_injectss live_arg_rep_injectss live_arg_abs_inversess P_ctor0' =
  HEADGOAL (rtac ctxt P_ctor0') THEN
  EVERY (@{map 4} (mk_nonuniform_induct_P_ctor_prem_tac ctxt fp_map_comps fp_map_idents fp_inj_maps)
    P_monos live_arg_abs_injectss live_arg_rep_injectss live_arg_abs_inversess);

fun mk_nonuniform_ind0_tac ctxt pre_pred_set P_ctor =
  HEADGOAL (rtac ctxt @{thm predicate1I} THEN'
    SELECT_GOAL (unfold_thms_tac ctxt (pre_pred_set :: @{thms o_apply Ball_def})) THEN'
    rtac ctxt P_ctor THEN' REPEAT_DETERM o clean_blast_tac ctxt);

fun mk_nonuniform_induct_Q_dtor0_tac ctxt Q_discsels pre_map_comps pre_map_idents fp_map_comps
    fp_map_idents live_arg_abs_inverses live_nesting_rel_eqs nn kk n pre_rel_def pre_abs_inverse
    dtor_ctor exhaust ctr_defs discss selss =
  let
    val extra_unfolds = pre_map_comps @ pre_map_idents @ fp_map_comps @ fp_map_idents @
      live_arg_abs_inverses @ @{thms id_def o_def};
  in
    unfold_thms_tac ctxt @{thms objectify_coind_prem} THEN
    HEADGOAL (Method.insert_tac ctxt Q_discsels THEN'
      mk_coinduct_discharge_prem_tac ctxt extra_unfolds live_nesting_rel_eqs nn kk n pre_rel_def
        pre_abs_inverse pre_abs_inverse dtor_ctor exhaust ctr_defs discss selss)
  end;

fun mk_nonuniform_induct_Q_dtor_tac ctxt live num_Xss live_arg_abs_injects live_arg_rep_inverses
    map_comps map_idents map_congs pre_rel_maps pre_rel_mono_strongs Q_dtor0 Q_monos =
  HEADGOAL (dtac ctxt Q_dtor0 THEN'
    SELECT_GOAL (unfold_thms_tac ctxt (id_apply :: pre_rel_maps)) THEN'
    EVERY' (@{map 3} (fn num_Xs => fn Q_mono => fn pre_rel_mono_strong =>
        etac ctxt pre_rel_mono_strong THEN'
        REPEAT_DETERM_N live o assume_tac ctxt THEN'
        REPEAT_DETERM_N num_Xs o
          (dtac ctxt ((Q_mono RS mp) |> Drule.rotate_prems ~1) THEN'
          REPEAT_DETERM_N num_Xs o
            (rtac ctxt @{thm injI} THEN'
             eresolve_tac ctxt live_arg_abs_injects THEN'
             asm_full_simp_tac (fold Simplifier.add_cong map_congs (ss_only
               (flat [live_arg_rep_inverses, map_comps, map_idents, @{thms o_apply}]) ctxt)))))
      num_Xss Q_monos pre_rel_mono_strongs));

end;
